/************************************************************
(C) Copyright 2006
Connect Tech Inc.
All rights reserved.
************************************************************/

/* ctibh.c -- terminal driver for Blue Heat/PCI */

/*
 * History
 *
 * Jul 10 2006 - cdore/stuartm
 *
 * First draft.
 */

static char *version = "1.0";
static char *revdate = "Jul 10 2006";
static char *drvname = "cbh";

#define __EXPLICIT_DEF_H__	/* Prevent ep.def.h from being included too early */

#include <kernel.h>         /* for SYSERR, etc. */
#include <pci_resource.h>
#include <sys/drm_errno.h>
#include <sys/ioctl.h>      /* for EVENP, etc. */
#include <ttymgr.h>         /* for struct ttystatics, etc. */
#include <dldd.h>
#include <errno.h>          /* for EINVAL, etc. */
#include <sys/file.h>       /* for FNDELAY */
#include <termio.h>         /* for CSIZE, CREAD, etc. */
#include <kern_proto.h>     /* for pgeterr(), etc. */

#include "ctibhinfo.h"  /* for struct ctibh_info (must come before ctibh.h)*/
#include "ctibh.h"
#include "ctibh_ioctl.h"


#if defined(__x86__)
#include <port_ops_x86.h>  /* for __inb(), etc. */
#define  EP_STATIC  static
/*
* Unfortunately, the skdb tty buffer support that was added to the x86
* version of kputc_x86.c was not propagated to the powerpc, so the
* following #define enables it for x86 only.
*/
#define __CTIBH_SKDB_BUFF__
#elif defined(__powerpc__)
#include <port_ops_ppc.h>  /* for __inb(), etc. */
#define  EP_STATIC
/*
* PowerPC has a devcon driver since you might be booting with either
* a serial port or a vt100 as your default console.
*/
#define __TTY_DEVCON__
#include <io.h>  /* for struct udriver_entry */
extern void devcon_register(void*, struct udriver_entry*, void (*)());
#endif  /* __powerpc__ */

#if defined(__arm__)
#include <port_ops_arm.h>
#define  EP_STATIC
#endif


struct ctibh_static;

struct ctibh_pstatic
{
	struct ctibh_static*  parent;
	__port_addr_t         uartp;
	int                   cts;  /* Clear To Send */
	unsigned char         fcr;
	int                   init;
	int                   dcd_check;
	struct ttystatics     channel;
};

struct ctibh_static {
	long                cb_len;
	struct drm_node_s*  drm_handle;

	__port_addr_t  va_bar1;

	int                    clock;
	unsigned char          num_ports;
	struct ctibh_pstatic*  ports;
};

extern void kkprintf_hook(int, int, void (*)(int, int));
extern void skdb_hook(int, int, struct skdb_vector*);
extern void (*skdb_entry_from_irq)(struct econtext*, int, int);
extern void kputc(int, int);  /* ctibh_kputc() is actually used */
static void ctibh_skdb(struct ctibh_pstatic*);
static int ctibh_kgetc(int);
void ctibh_skdb_hook(int, int);

#define ctibh_kputc  kputc  /* kputc() in the kernel */


#define MAX_COUNT  100


extern int    conv_vb[];  /* defined in ttymgr.c */
extern int    conv_bv[];  /* defined in ttymgr.c */
extern int    skdb_tty_head;
extern int    skdb_tty_tail;
extern char   skdb_tty_buff[];
extern char   skdb_key_serial;
extern st_t*  currtptr;


#if defined(__TTY_DEVCON__)
extern int ctibhopen(struct ctibh_static*, int, struct file*);
extern int ctibhclose(struct ctibh_static*, struct file*);
extern int ctibhread(struct ctibh_static*, struct file*, char*, int);
extern int ctibhwrite(struct ctibh_static*, struct file*, char*, int);
extern int ctibhselect(struct ctibh_static*, struct file*, int, struct sel*);
extern int ctibhioctl(struct ctibh_static*, struct file*, int, char*);
extern struct ctibh_static* ctibhinstall(struct ctibh_info*);
extern int ctibhuninstall(struct ctibh_static*);

static struct udriver_entry  ctibh_funcs = {
	0, NULL,
	ctibhopen, ctibhclose, ctibhread, ctibhwrite, ctibhselect, ctibhioctl,
	ctibhinstall, ctibhuninstall, NULL
};
#endif  /* __TTY_DEVCON__ */

#if defined(__GNUC__)
#define __UNUSED_ARG(x, y...)  x y __attribute__ ((unused))
#else
#define __UNUSED_ARG(x)  x
#endif

#if defined(uLYNX)  /* the following include file is necessary for Galaxy */
#include <ep.def.h>  /* PLEASE DO NOT place any extern's after this line */
#endif  /* Thank you. - The Management */

static int
ctibh_set_baud(
	struct ctibh_pstatic*  p,
	int                    ispeed,
	int                    ospeed
	)
{
	__port_addr_t  uartp = p->uartp;
	int            baud;
	int            divisor;
	int            ps;
	int            c;

	debug(("%s: enter\n", __FUNCTION__));

	switch (ospeed) {
		case B0:      baud = 0;      break;
		case B50:     baud = 50;     break;
		case B75:     baud = 75;     break;
		case B110:    baud = 110;    break;
		case B134:    baud = 134;    break;
		case B150:    baud = 150;    break;
		case B200:    baud = 200;    break;
		case B300:    baud = 300;    break;
		case B600:    baud = 600;    break;
		case B1200:   baud = 1200;   break;
		case B1800:   baud = 1800;   break;
		case B2000:   baud = 2000;   break;
		case B2400:   baud = 2400;   break;
		case B4800:   baud = 4800;   break;
		case B7200:   baud = 7200;   break;
		case B9600:   baud = 9600;   break;
		case B19200:  baud = 19200;  break;
		case B38400:  baud = 38400;  break;
		case B57600:  baud = 57600;  break;
		default:
		case B115200: baud = 115200; break;
#if 0
		case EXTA:
		case EXTB:
#endif
	}
	divisor = p->parent->clock / 16 / baud;
	if (ospeed == B0) {
		/*
			According to POSIX.1b, section 7.2.1.2, setting the output baud
			rate to B0 causes the modem control lines to be disconnected.
		*/
		__port_andb(uartp + U_MCR, ~(MCR_DTR | MCR_RTS | MCR_OT2));
	}

	debug(("%s: set baud idx=%d baud=%d div=%d\n", __FUNCTION__, ospeed, baud, divisor));

	disable(ps);
	c = __inb(uartp + U_LCR);			/* save orig line control register */

	__outb(uartp + U_LCR, c | LCR_DLAB);	/* select DLAB */

	__outb(uartp + U_DLL, divisor % 256);		/* write low divisor */
	__outb(uartp + U_DLM, divisor / 256);		/* write high divisor */

	__outb(uartp + U_LCR, c);				/* deselect DLAB */
	restore(ps);
	
	debug(("%s: leave\n", __FUNCTION__));
	return OK;
}

static int
ctibh_setfifo(
	struct ctibh_pstatic*  p,
	int                    fifo_enable,
	unsigned char          rx_trg,
	unsigned char          tx_trg,
	int                    rx_reset,
	int                    tx_reset
	)
{
	int            ps;
	unsigned char  fcr = 0;

	if(fifo_enable)
		fcr |= FCR_EN;

	switch(rx_trg)
	{
	case FCR_RTL8:
	case FCR_RTL16:
	case FCR_RTL56:
	case FCR_RTL60:
		fcr |= rx_trg;
		break;
	default:
		debug(("%s: invalid rx trigger level (0x%X)\n", __FUNCTION__, rx_trg));
		return SYSERR;
	}

	switch(tx_trg)
	{
	case FCR_TTL8:
	case FCR_TTL16:
	case FCR_TTL32:
	case FCR_TTL56:
		fcr |= tx_trg;
		break;
	default:
		debug(("%s: invalid tx trigger level (0x%X)\n", __FUNCTION__, tx_trg));
		return SYSERR;
	}

	if(rx_reset)
		fcr |= FCR_RFR;

	if(tx_reset)
		fcr |= FCR_TFR;

	disable(ps);
	__outb(p->uartp + U_FCR, fcr);
	p->fcr = fcr;
	restore(ps);

	return(OK);
}

/*
	ctibh_seth -- set hardware dependant tty settings
*/
static void
ctibh_seth(
	struct ttystatics*     s,
	struct sgttyb*         sg,
	struct ctibh_pstatic*  p,
	int                    firsttime,
	int                    tflags
	)
{
	__port_addr_t  uartp = p->uartp;
	unsigned char  lcr;

	debug(("%s: enter\n", __FUNCTION__));

	swait(&s->hsem, -1);
	if (firsttime || (sg->sg_flags & (EVENP|ODDP)) != (tflags & (EVENP|ODDP))) {
		if (sg->sg_flags & EVENP) {
			if (sg->sg_flags & ODDP) {	/* both even and odd 7 bit no parity */
				__outb(uartp + U_LCR, LCR_WL7 | LCR_STB_1);
			}
			else {
				__outb(uartp + U_LCR, LCR_WL7 | LCR_PEN | LCR_EPS | LCR_STB_1);
			}
		}
		else if (sg->sg_flags & ODDP) {
			__outb(uartp + U_LCR, LCR_WL7 | LCR_PEN | LCR_STB_1);
		}
		else {
			__outb(uartp + U_LCR, LCR_WL8 | LCR_STB_1);
		}
	}

	if (s->sg.sg_ispeed != sg->sg_ispeed || s->sg.sg_ospeed != sg->sg_ospeed) {
		if (!ctibh_set_baud(p, sg->sg_ispeed, sg->sg_ospeed)) {
			if (s->sg.sg_ospeed == B0 && sg->sg_ospeed != B0) {
				/*
					Changing output speed from B0 to non-B0, so it is
					necessary to reconnect the modem control lines.
				*/
				__port_orb(uartp + U_MCR, (MCR_DTR | MCR_RTS | MCR_OT2));
			}
			s->sg.sg_ispeed = sg->sg_ispeed;
			s->sg.sg_ospeed = sg->sg_ospeed;
		}
	}

	/*
	* in order to enable the tx trigger level, the enhanced functions
	* bit must be enabled
	*/
	lcr = __inb(p->uartp + U_LCR);
	__outb(p->uartp + U_LCR, lcr | 0xbf);
	__outb(p->uartp + U_EFR, EFR_EFB);
	__outb(p->uartp + U_LCR, lcr);
	ctibh_setfifo(p, 1, FCR_RTL56, FCR_TTL32, 1, 1);

	ssignal(&s->hsem);

	debug(("%s: leave\n", __FUNCTION__));
}

/*
	ctibh_trans_en -- transmitter enable. The only routine called by the tty manager
*/
static void
ctibh_trans_en(
	char*  arg
	)
{
	struct ctibh_pstatic*  port_statics;
	int                    ps;

	debug(("%s: enter\n", __FUNCTION__));

	port_statics = (struct ctibh_pstatic*)arg;
	/*
		The transmitter should not be enabled if output hardware
		flow-control is being performed and CTS (in the MSR) is not set.
		Keep in mind that CRTSCTS means both input and output hardware
		flow-control (Linux compatible).
	*/

	if (!(port_statics->channel.tm.c_cflag & V_CRTSCTS) ||
		(__inb(port_statics->uartp + U_MSR) & MSR_CTS))
	{
		/*
			Either output hardware flow-control is not being performed, or
			CTS (in the MSR) is set, so it's OK to enable the transmitter.
		*/
		disable(ps);
		__port_orb(port_statics->uartp + U_IER, IER_TRE);
		restore(ps);
	}
	debug(("%s: leave\n", __FUNCTION__));
}

/*
	ctibh_duart_int - interrupt routine.
*/
static inline void
ctibh_duart_int(
		struct ctibh_pstatic*  p  /* ptr to cti port structure */
		)
{
	__port_addr_t uartp;
	struct ttystatics *s;			/* ptr to static struct for a chan */
	int status;
	int data;
	unsigned char c;				/* make unsigned to aid debugging */
	int got_break = 0;				/* added for inbreak check */
	int parity_error = 0;			/* added for parity check */

	uartp = p->uartp;
	s = &p->channel;

	for (;;) {
		if ((status = __inb(uartp + U_IIR)) & IIR_PEN) { /* no intr pending */
			break;
		}
		switch (status & IIR_IID) {
			case IIR_RLS:				/* receiver error */
				if ((data = __inb(uartp + U_LSR)) & LSR_BI) {
					/* Break interrupt */
					got_break = 1;
					tmgr_ex(s, EX_BREAK, 0);
				}
				if (data & (LSR_PE | LSR_FE)) {
					/* Parity error or framing error */
					if (s->xflags1 & X1_ENABLE_INPUT_PC) {
						parity_error = 1;
					}
				}
				if ((data & LSR_BI) || !(data & (LSR_PE | LSR_FE))) {
					/* Flush the receive buffer register */
					data = __inb(uartp + U_RBR);
				}
				break;
			case IIR_RDA:				/* receiver data ready */
				/* TODO: Loop this on LSR's DR bit */
				data = __inb(uartp + U_RBR);
				if (skdb_entry_from_irq && __ctibh_is_skdb_port(uartp)
						&& data == skdb_key_serial) {
					ctibh_skdb(p);
					continue;
				}

				/*
					The ttydrvr receives a 0 on transition
					from break on to break off.
				*/
				if (!data && (s->t_line != SLIPDISC)) {
					if (got_break == 1) {
						got_break = 0;
						break;			/* dont send character on */
					}
				}
				if (!parity_error) {
					if (tmgr_rx(s, data)) {
						/* enable transmit interrupt */
						__port_orb(uartp + U_IER, IER_TRE);
					}
				}
				else {	/* parity error */
					tmgr_ex(s, EX_PARITY, data);
					parity_error = 0;
					/* enable transmit interrupt */
					__port_orb(uartp + U_IER, IER_TRE);
				}
				break;
			case IIR_TRE:				/* transmitter empty */
				if (p->dcd_check && !p->cts) {
					/* do not send because "No CTS" */
					break;
				}
				/* TODO: loop this based on tx fifo trigger */
				if (tmgr_tx(s, (char *)&c)) {			/* call manager */
					/* disable transmit interrupt */
					__port_andb(uartp + U_IER, ~IER_TRE);
				}
				else {
					__outb(uartp + U_THR, c);
				}
				break;
			case IIR_MS:
				/* flow control hardware handshake */
				data = __inb(uartp + U_MSR);
				/*
					This case should only occur during an open when
					the file is configured as blocking, OR when we're
					doing hardware flow-control.  Keep in mind that
					CRTSCTS means both input and output hardware
					flow-control (Linux compatible).  The dcd_check
					is for the open routine to block appropriately.
				*/
				if ((data & MSR_DCTS) && (p->dcd_check
					|| (s->tm.c_cflag & (V_CRTSCTS | V_CRTSXOFF)))) {
					if (!(data & MSR_CTS)) {	/* DCE is not ready */
						p->cts = 0;
					}
					else {
						p->cts = 1;
						/* enable transmit interrupt */
						__port_orb(uartp + U_IER, IER_TRE);
					}
				}
				if (data & MSR_DDCD) {
					/* MSR_DDCD set => MSR_DCD was raised or dropped */
					tmgr_ex(s, (data & MSR_DCD) ? EX_CARRIER : EX_HUP, 0);
				}
				break;
			case IIR_CTI:
				/* missed too many (>4) interrupts and timed out */
				while ((__inb(uartp + U_LSR) & LSR_DR)) {
					data = __inb(uartp + U_RBR);
					tmgr_rx(s, data);
				}
				break;
		}
	}
}

static void
ctibh_int(
	struct ctibh_static*  board  /* ptr to cti board structure */
	)
{
	unsigned char stat_port;
	int i;

	isr_debug(("%s: enter\n", __FUNCTION__));

	stat_port = __inb(board->va_bar1 + BH_STATUS_OFFSET);
	i = 0;
	while (stat_port) {
		isr_debug(("%s: stat:%02x\n", __FUNCTION__, stat_port));
		if (stat_port & 0x1) {
			isr_debug(("%s: i:%d\n", __FUNCTION__, i));
			ctibh_duart_int(&board->ports[i]);
		}
		stat_port >>= 1;
		i++;
		if (!stat_port) {
			stat_port = __inb(board->va_bar1 + BH_STATUS_OFFSET);
			i = 0;
		}
	}
	isr_debug(("%s: leave\n", __FUNCTION__));
}

static struct ctibh_static*
ctibh_build_statics(
	struct ctibh_info*  info
	)
{
	struct drm_node_s*  drm_handle;
	struct ctibh_static*  ret = NULL;
	int                   rv;
	int                   busno, devno, funcno;
	int vendor, dev;
	unsigned char         nports;
	unsigned char         i;
	long                  cbStatics;
	unsigned int          cmdstat;
	int clock;

	rv = drm_get_handle(PCI_BUSLAYER, V3_VENDORID, 2, &drm_handle);
	if (rv != DRM_OK)
		rv = drm_get_handle(PCI_BUSLAYER, V3_VENDORID, 1, &drm_handle);
	if (rv != DRM_OK) {
		debug(("%s: failed to get handle\n", __FUNCTION__));
		goto failure;
	}

	rv = drm_device_read(drm_handle, PCI_RESID_BUSNO, 0, 0, &busno);
	if (rv != DRM_OK) {
		debug(("%s: failed to read busno (%d)\n", __FUNCTION__, rv));
		goto failure;
	}
	rv = drm_device_read(drm_handle, PCI_RESID_DEVNO, 0, 0, &devno);
	if (rv != DRM_OK) {
		debug(("%s: failed to read devno (%d)\n", __FUNCTION__, rv));
		goto failure;
	}
	rv = drm_device_read(drm_handle, PCI_RESID_FUNCNO, 0, 0, &funcno);
	if (rv != DRM_OK) {
		debug(("%s: failed to read funcno (%d)\n", __FUNCTION__, rv));
		goto failure;
	}
	debug(("%s: device location is bus:dev:func=%d:%d:%d\n", __FUNCTION__, busno, devno, funcno));

	rv = drm_device_read(drm_handle, PCI_RESID_SUBSYSVID, 0, 0, &vendor);
	if (rv != DRM_OK) {
		debug(("%s: failed to read subsys (%d)\n", __FUNCTION__, rv));
		goto failure;
	}
	if (vendor != CTIBH_VENDORID) {
		debug(("%s: wrong vendor id (%d)\n", __FUNCTION__, vendor));
		goto failure;
	}

	rv = drm_device_read(drm_handle, PCI_RESID_SUBSYSID, 0, 0, &dev);
	if (rv != DRM_OK) {
		debug(("%s: failed to read subsys (%d)\n", __FUNCTION__, rv));
		goto failure;
	}

	switch (dev) {
		default:
		case 0x0e:
		case 0x01: nports = 8; clock = 22118400; break;
		case 0x02: nports = 4; clock = 22118400; break;
		case 0x03: nports = 2; clock = 22118400; break;
		case 0x04: nports = 8; clock = 14745600; break;
		case 0x05: nports = 8; clock = 14745600; break;
		case 0x06: nports = 4; clock = 14745600; break;
		case 0x07: nports = 4; clock = 14745600; break;
		case 0x08: nports = 2; clock = 14745600; break;
		case 0x09: nports = 8; clock = 14745600; break;
		case 0x0a: nports = 8; clock = 14745600; break;
		case 0x0b: nports = 4; clock = 14745600; break;
		case 0x0c: nports = 2; clock = 20000000; break;
		case 0x0d: nports = 2; clock = 14745600; break;
		case 0x0f: nports = 8; clock = 19692800; break;
		case 0x10: nports = 8; clock = 16240000; break;
	}
	debug(("%s: nports=%d clock=%u\n", __FUNCTION__, nports, clock));

	cbStatics = (long)sizeof(struct ctibh_static) + sizeof(struct ctibh_pstatic) * nports;
	ret = (struct ctibh_static*)sysbrk(cbStatics);
	if (ret == NULL) {
		debug(("%s: failed to allocate static structure\n", __FUNCTION__));
		goto failure;
	}
	
	ret->cb_len = cbStatics;
	ret->drm_handle = drm_handle;
	ret->va_bar1 = 0;
	ret->clock = clock;
	ret->num_ports = nports;
	ret->ports = (struct ctibh_pstatic*)((size_t)&(ret->ports) + sizeof(ret->ports));

	rv = drm_map_resource(drm_handle, PCI_RESID_BAR1, &ret->va_bar1);
	if(rv != DRM_OK) {
		debug(("%s: failed to map bar1 (%d)\n", __FUNCTION__, rv));
		goto failure;
	}
	debug(("%s: mapped = 0x%08X\n", __FUNCTION__, ret->va_bar1));

	for(i = 0; i < ret->num_ports; i++) {
		ret->ports[i].parent = ret;
	}

	rv = drm_device_read(drm_handle, PCI_RESID_CMDREG, 0, 0, &cmdstat);
	if(rv != DRM_OK) {
		debug(("%s: failed to read PCI cmd reg (%d)\n", __FUNCTION__, rv));
		goto failure;
	}
	debug(("%s: PCI cmd reg = 0x%X\n", __FUNCTION__, cmdstat));
	cmdstat &= 0xffff;
	cmdstat |= PCI_IO_ENABLE;
	rv = drm_device_write(drm_handle, PCI_RESID_CMDREG, 0, 0, &cmdstat);
	if(rv != DRM_OK) {
		debug(("%s: failed to write PCI cmd reg (%d)\n", __FUNCTION__, rv));
		goto failure;
	}

	cprintf("%s: BlueHeat %d port %d clock (%02d:%02d.%d)\n", drvname, nports, clock, busno, devno, funcno);

	goto done;

failure:
	if(ret != NULL) {
		for(i = 0; i < ret->num_ports; i++) {
			tmgr_uninstall(&ret->ports[i].channel);
		}
		if(ret->va_bar1 != 0)
			drm_unmap_resource(drm_handle, PCI_RESID_BAR1);
		sysfree((char*)ret, ret->cb_len);
		ret = NULL;
	}

done:
	return(ret);
}

char*
ctibhinstall(
	struct ctibh_info*  info  /* device information */
	)
{
	struct ctibh_static*  dev_statics;
	unsigned char         i;

	cprintf("%s: BlueHeat driver v%s %s\n", drvname, version, revdate);

	debug(("%s: enter (0x%X)\n", __FUNCTION__, info));

	dev_statics = ctibh_build_statics(info);
	if(dev_statics == NULL)
		dev_statics = (struct ctibh_static*)SYSERR;
	else {
		/* init uarts */
		for(i = 0; i < dev_statics->num_ports; i++) {
			dev_statics->ports[i].uartp = dev_statics->va_bar1 + (i * 8);
			cprintf("%s: Port %d at 0x%08x\n", drvname, i, dev_statics->ports[i].uartp);
			dev_statics->ports[i].cts = 0;

			tmgr_install(&dev_statics->ports[i].channel, &info->sg, 0, ctibh_trans_en, (char*)&dev_statics->ports[i]);
			dev_statics->ports[i].channel.hsem = 1;

			ctibh_seth(&(dev_statics->ports[i].channel), (struct sgttyb*)&info->sg, &dev_statics->ports[i], 1, 0);
			/* the device is not open yet, so disconnect the modem control lines */
			__port_andb(dev_statics->ports[i].uartp + U_MCR, ~(MCR_DTR | MCR_RTS | MCR_OT2));

			dev_statics->ports[i].init = 0;
#if defined(__TTY_DEVCON__)
			/* register the entrypoints with devcon driver in case it is console */
			devcon_register((void*)&dev_statics->ports[i], &ctibh_funcs, ctibh_skdb_hook);
#endif
		}

		/* register interrupt routine */
		drm_register_isr(dev_statics->drm_handle, ctibh_int, dev_statics);

		for(i = 0; i < dev_statics->num_ports; i++) {
			/* enable the modem status interrupts */
			__port_orb(dev_statics->ports[i].uartp + U_IER, IER_MS);
		}

	}

	debug(("%s: leave\n", __FUNCTION__));
	return (char*)dev_statics;
}

/*
	ctibh_ttyseth_pv -	Set hardware dependant tty settings using either the
					POSIX termios structure or the System V termio structure.
						sysv = 0	=> (POSIX):		struct termios *sg
						sysv != 0	=> (System V):	struct termio *sg
*/
static void
ctibh_seth_pv(
		struct ttystatics*     s,
		struct termios*        sg,     /* type depends on value of sysv */
		struct ctibh_pstatic*  p,  /* base address of uart */
		int                    sysv    /* (0 => struct termios *sg) (~0 => struct termio *sg) */
		)
{
 	__port_addr_t  uartp = p->uartp;
	int            speed;
	unsigned char  csize;
	unsigned char  cread;
	tcflag_t       c_cflag;    /* local storage for c_cflag */
	unsigned int   lcr_flags;
	int            ps;
	int            index;      /* user set baudrate, support for 57600 & 115200 */

	debug(("%s: enter\n", __FUNCTION__));

	if (sysv) {
		/* sg points to a System V termio structure */
		csize = ((struct termio *)sg)->c_cflag & CSIZE;
		cread = ((struct termio *)sg)->c_cflag & CREAD;
		c_cflag = ((struct termio *)sg)->c_cflag;
	}
	else {
		/* sg points to a POSIX termios structure */
		csize = sg->c_cflag & CSIZE;
		cread = sg->c_cflag & CREAD;
		c_cflag = sg->c_cflag;
	}
	index = (c_cflag & V_CBAUDBASE) + ((c_cflag & V_CBAUDEX) ? V_EXTBASE : 0);
	if (index <= V_EXTB) {
		speed = conv_vb[index];
	}
	else {
		cprintf("ctibh: baudrate OUT OF RANGE set to B0\n");
		speed = B0;
	}

	swait(&s->hsem, -1);
	switch (csize) {
		case (CS5):	lcr_flags = LCR_WL5; break;
		case (CS6):	lcr_flags = LCR_WL6; break;
		case (CS7):	lcr_flags = LCR_WL7; break;
		case (CS8):	lcr_flags = LCR_WL8; break;
		default:	lcr_flags = LCR_WL7; break;
	}
	if (s->xflags1 & X1_ENABLE_PC) {
		lcr_flags |= LCR_PEN;
	}
	if (s->sg.sg_flags & EVENP) {
		lcr_flags |= LCR_EPS;
	}

	if (s->xflags1 & X1_STOP_2) {
		lcr_flags |= LCR_STB_2;
	}
	else {
		lcr_flags |= LCR_STB_1;
	}

	/* Set Line Control Register */
	disable(ps);
	__outb(uartp + U_LCR, lcr_flags);
	restore(ps);

	if (s->sg.sg_ispeed != speed || s->sg.sg_ospeed != speed) {
		if (!ctibh_set_baud(p, speed, speed)) {
			if (s->sg.sg_ospeed == B0 && speed != B0) {
				/*
					Changing output speed from B0 to non-B0, so it is
					necessary to reconnect the modem control lines.
				*/
				__port_orb(uartp + U_MCR, (MCR_DTR | MCR_RTS | MCR_OT2));
			}
			s->sg.sg_ispeed = speed;
			s->sg.sg_ospeed = speed;
		}
		/*
			The following code is not affected by the new support for
			57600 and 115200 because ispeed is not a bitmask, but an
			index to get the old baudrate bitmask out of conv_bv[].
		*/
		s->tm.c_cflag = (s->tm.c_cflag & ~V_CBAUD)
				| conv_bv[(int)s->sg.sg_ispeed];
	}

	if (cread) {		/* Enable receiver */
		disable(ps);
		__port_orb(uartp + U_IER, (IER_RDA | IER_RLS));
		restore(ps);
	}
	else {				/* Disable receiver */
		disable(ps);
		__port_andb(uartp + U_IER, ~(IER_RDA | IER_RLS));
		restore(ps);
	}
	ssignal(&s->hsem);
	debug(("%s: leave\n", __FUNCTION__));
}

int
ctibhuninstall(
	struct ctibh_static*  dev_statics
	)
{
	int            ret = OK;
	unsigned char  i;
	int            rv;

	debug(("%s: enter (0x%X)\n", __FUNCTION__, dev_statics));

	if(dev_statics != NULL) {
		for(i = 0; i < dev_statics->num_ports; i++) {
			__outb(dev_statics->ports[i].uartp + U_IER, 0);
			tmgr_uninstall(&(dev_statics->ports[i].channel));
		}
		rv = drm_unmap_resource(dev_statics->drm_handle, PCI_RESID_BAR1);
		if(rv != DRM_OK) {
			debug(("%s: failed to unmap bar1 (%d)\n", __FUNCTION__, rv));
			ret = SYSERR;
		}
		rv = drm_unregister_isr(dev_statics->drm_handle);
		if(rv != DRM_OK) {
			debug(("%s: failed to unregister isr (%d)\n", __FUNCTION__, rv));
			ret = SYSERR;
		}
		rv = drm_free_handle(dev_statics->drm_handle);
		if(rv != DRM_OK) {
			debug(("%s: failed to free the device handle (%d)\n", __FUNCTION__, rv));
			ret = SYSERR;
		}
		sysfree((char*)dev_statics, (long)sizeof(struct ctibh_static));
	}
	else
		ret = SYSERR;

	debug(("%s: leave\n", __FUNCTION__));
	return(ret);
}

/*
	ctibh_xmit_quies - Wait for the transmitter to be quiescent.
*/
void
ctibh_xmit_quies(
	__port_addr_t  uartp
	)
{
	debug(("%s: enter\n", __FUNCTION__));
	while((__inb(uartp + U_LSR) & (LSR_TRE | LSR_TSE))
		!= (LSR_TRE | LSR_TSE))
	{
		if(tswait(0, SEM_SIGABORT, 1) == SYSERR) {
			/* interrupted by a signal */
			break;
		}
	}
	debug(("%s: leave\n", __FUNCTION__));
}

int
ctibhclose(
	struct ctibh_static*  p,
	struct file*          f
	)
{
	int            minor_no;
	__port_addr_t  uartp;

	debug(("%s: enter (0x%X,0x%X)\n", __FUNCTION__, p, f));

	minor_no = minor(f->dev);
	if((minor_no < 0) || (minor_no >= p->num_ports)) {
		debug(("%s: minor number out of range (%d)\n", __FUNCTION__, minor_no));
		return(SYSERR);
	}

	tmgr_close(&(p->ports[minor_no].channel), f);
	uartp = p->ports[minor_no].uartp;
	ctibh_xmit_quies(uartp);
	/*
		According to the POSIX Test Suite (DC:settable:054),
		if HUPCL and CLOCAL are both clear, then the last close
		shall not cause the modem control lines to be disconnected.
	*/
	if ((p->ports[minor_no].channel.xflags1 & (X1_HUP_CLOSE | X1_MODEM_OK)) != X1_MODEM_OK) {
		/* HUPCL is set or CLOCAL is set */
		__port_andb(uartp + U_MCR, ~(MCR_DTR | MCR_RTS | MCR_OT2));
	}
	__outb(uartp + U_IER, 0);  /* disable serial interrupts */
	p->ports[minor_no].channel.flags &= ~ACTIVE;
	p->ports[minor_no].init = 0;

	debug(("%s: leave\n", __FUNCTION__));
	return(OK);
}

static int
ctibh_hardware_handshake(
	struct ctibh_static*  p,
	struct file*          f,
	struct ttystatics*    s
	)
{
	int            minor_no;
	int            counter = tickspersec / 10;
	__port_addr_t  uartp;

	debug(("%s: enter\n", __FUNCTION__));

	minor_no = minor(f->dev);
	if((minor_no < 0) || (minor_no >= p->num_ports)) {
		debug(("%s: minor number out of range (%d)\n", __FUNCTION__, minor_no));
		return(SYSERR);
	}

	uartp = p->ports[minor_no].uartp;
	__port_orb(uartp + U_MCR, MCR_DTR);

	p->ports[minor_no].dcd_check = (s->xflags1 & X1_MODEM_OK) && !(f->access_mode & FNDELAY);

	if (p->ports[minor_no].dcd_check) {
		while (!(__inb(uartp + U_MSR) & MSR_DCD) && counter-- > 0) {
			if (tswait(0, SEM_SIGABORT, 1) != TSWAIT_TIMEDOUT) {
				p->ports[minor_no].dcd_check = 0;
				ctibh_trans_en((char*)&(p->ports[minor_no]));
				debug(("%s: leave\n", __FUNCTION__));
				return SYSERR;
			}
			__port_orb(uartp + U_MCR, MCR_DTR);
		}
	}
	p->ports[minor_no].dcd_check = 0;
	ctibh_trans_en((char*)&(p->ports[minor_no]));
	debug(("%s: leave\n", __FUNCTION__));
	return OK;
}

int
ctibhopen(
	struct ctibh_static*  p,
	int                   dev,
	struct file*          f
	)
{
	int                 minor_no;
	int                 out;
	struct ttystatics*  s;
	__port_addr_t       uartp;
	int                 data;
	int                 i;

	debug(("%s: enter (0x%X,0x%X,0x%X)\n", __FUNCTION__, p, dev, f));

	minor_no = minor(f->dev);
	if((minor_no < 0) || (minor_no >= p->num_ports)) {
		debug(("%s: minor number out of range (%d)\n", __FUNCTION__, minor_no));
		return(SYSERR);
	}

	s = &(p->ports[minor_no].channel);
	if ((out = tmgr_open(s, f)) == SYSERR) {
		debug(("%s: tmgr_open failed (%d)\n", __FUNCTION__, out));
		return SYSERR;
	}

	s->flags |= ACTIVE;
	if (s->sg.sg_ospeed != B0) {
		/*
			The output baud rate is not B0,
			so it's OK to assert modem control lines.
		*/
		if (ctibh_hardware_handshake(p, f, s) == SYSERR) {
			if (out) {
				ctibhclose(p, f);
			}
			debug(("%s: leave\n", __FUNCTION__));
			return SYSERR;
		}
	}
	if (out && 0 == p->ports[minor_no].init) {				/* first open */
		p->ports[minor_no].init = 1;
		uartp = p->ports[minor_no].uartp;
		__port_orb(uartp + U_MCR, MCR_OT2 | MCR_RTS);
		__outb(uartp + U_LCR, LCR_WL8 | LCR_STB_1);
		data = __inb(uartp + U_IIR);
		data = __inb(uartp + U_MSR);
		/* flush hardware buffer before enabling interrupts */
		for (i = 0; i < MAX_COUNT; i++) {
			if (!((data = __inb(uartp + U_LSR)) & IER_RDA)) {
				break;				/* no more fake data */
			}
			data = __inb(uartp + U_RBR);
		}
		__outb(uartp + U_IER, IER_RDA | IER_TRE | IER_RLS | IER_MS);
	}

	debug(("%s: leave\n", __FUNCTION__));
	return OK;
}

int
ctibhread(
	struct ctibh_static*  p,
	struct file*          f,
	char*                 buff,
	int                   count
	)
{
	int  minor_no;
	int  ret;

	debug(("%s: enter (0x%X,0x%X,0x%X,0x%X)\n", __FUNCTION__, p, f, buff, count));

	minor_no = minor(f->dev);
	if((minor_no < 0) || (minor_no >= p->num_ports)) {
		debug(("%s: minor number out of range (%d)\n", __FUNCTION__, minor_no));
		return(SYSERR);
	}

	ret = tmgr_read(&(p->ports[minor_no].channel), f, buff, count);
	debug(("%s: leave\n", __FUNCTION__));
	return ret;
}

int
ctibhwrite(
	struct ctibh_static*  p,
	struct file*          f,
	char*                 buff,
	int                   count
	)
{
	int  minor_no;
	int  ret;

	debug(("%s: enter (0x%X,0x%X,0x%X,0x%X)\n", __FUNCTION__, p, f, buff, count));

	minor_no = minor(f->dev);
	if((minor_no < 0) || (minor_no >= p->num_ports)) {
		debug(("%s: minor number out of range (%d)\n", __FUNCTION__, minor_no));
		return(SYSERR);
	}

	ret = tmgr_write(&(p->ports[minor_no].channel), f, buff, count);
	debug(("%s: leave\n", __FUNCTION__));
	return ret;
}

/*
 * translation from ioctl bit numberings to UART bit numberings
 */
struct ioctlbitmapping {
	int ioctl_bit;
	int uart_bit;
};

static struct ioctlbitmapping mcr_map[] = {
	{TIOCM_DTR, MCR_DTR},
	{TIOCM_RTS, MCR_RTS},
	{TIOCM_OUT1, MCR_OT1},
	{TIOCM_OUT2, MCR_OT2},
	{TIOCM_LE, MCR_LOOP},
	{-1, -1}
};

static struct ioctlbitmapping msr_map[] = {
	{TIOCM_CTS, MSR_CTS},
	{TIOCM_DSR, MSR_DSR},
	{TIOCM_CD, MSR_DCD},
	{TIOCM_RI, MSR_RI},
	{-1, -1}
};

/* translate from ioctl bits to UART bits */
static int
ctibh_ioctl2mcr(
	int  ioctlbits
	)
{
	int uart_mcr = 0;
	struct ioctlbitmapping *mapper;
	debug(("%s: enter\n", __FUNCTION__));
	for (mapper = mcr_map; mapper->ioctl_bit != -1; mapper++) {
		if (ioctlbits & mapper->ioctl_bit) {
			uart_mcr |= mapper->uart_bit;
		}
	}
	debug(("%s: leave\n", __FUNCTION__));
	return uart_mcr;
}

static int
ctibh_ioctl2msr(
	int  ioctlbits
	)
{
	int uart_msr = 0;
	struct ioctlbitmapping *mapper;
	debug(("%s: enter\n", __FUNCTION__));
	for (mapper = msr_map; mapper->ioctl_bit != -1; mapper++) {
		if (ioctlbits & mapper->ioctl_bit) {
			uart_msr |= mapper->uart_bit;
		}
	}
	debug(("%s: leave\n", __FUNCTION__));
	return uart_msr;
}

static int
ctibh_calc_baud(struct ctibh_pstatic *p, int baud)
{
	int divisor;
	int error;
	int base_baud = p->parent->clock / 16;
	__port_addr_t uartp = p->uartp;
	unsigned char c;
	int ps;

	if (baud < 0 || baud > base_baud)
		goto error_out;

	if (!baud) {
		divisor = 0;
		/*
			According to POSIX.1b, section 7.2.1.2, setting the output baud
			rate to B0 causes the modem control lines to be disconnected.
		*/
		__port_andb(uartp + U_MCR, ~(MCR_DTR | MCR_RTS | MCR_OT2));
	} else {
		divisor = base_baud / baud;

		if (base_baud % baud) {
			error = base_baud / divisor;
			error *= 100;
			error /= baud;
			error -= 100;
			if (error < 0)
				error *= -1;
			if (error > 2)
				goto error_out;
		}
	}

	debug(("%s: set baud baud=%d div=%d\n", __FUNCTION__, baud, divisor));

	disable(ps);
	c = __inb(uartp + U_LCR);			/* save orig line control register */

	__outb(uartp + U_LCR, c | LCR_DLAB);	/* select DLAB */

	__outb(uartp + U_DLL, divisor % 256);		/* write low divisor */
	__outb(uartp + U_DLM, divisor / 256);		/* write high divisor */

	__outb(uartp + U_LCR, c);				/* deselect DLAB */
	restore(ps);

	return OK;

error_out:
	return SYSERR;
}

int
ctibhioctl(
	struct ctibh_static*  b,
	struct file*          f,
	int                   com,
	char*                 arg
	)
{
	int                      minor_no;
	struct ctibh_pstatic*    p;
	__port_addr_t            uartp;
	struct ttystatics*       s;
	struct ioctlbitmapping*  mapper;
	int                      out;
	int                      in;
	int                      tflags;
	int                      uart_mcr;
	int                      uart_msr;
	int                      ps;

	debug(("%s: enter (0x%X,0x%X,0x%X,0x%X)\n", __FUNCTION__, b, f, com, arg));

	minor_no = minor(f->dev);
	if((minor_no < 0) || (minor_no >= b->num_ports)) {
		debug(("%s: minor number out of range (%d)\n", __FUNCTION__, minor_no));
		return(SYSERR);
	}

	p = &(b->ports[minor_no]);
	s = &p->channel;

	tflags = s->sg.sg_flags;
	uartp = p->uartp;

	/* initial values for modem status and control registers */
	uart_msr = 0;
	uart_mcr = 0;

	/*
	 * Custom ioctls for Connect Tech cards. Fall through on ioctls
	 * that we don't care about so they are handled normally.
	 */
	switch (com) {
		case CBH_SET_BAUD:
			return ctibh_calc_baud(p, (int)arg);
	}

	if (!(out = tmgr_ioctl(s, f, com, arg))) {
		switch (com) {
			case TIOCSDTR:		/* set data terminal ready */
				__port_orb(uartp + U_MCR, MCR_DTR);
				return OK;
			case TIOCCDTR:		/* clear data terminal ready */
				__port_andb(uartp + U_MCR, ~MCR_DTR);
				return OK;
			case TIOCSRTS:		/* set request to send */
				__port_orb(uartp + U_MCR, MCR_RTS);
				return OK;
			case TIOCCRTS:		/* clear request to send */
				__port_andb(uartp + U_MCR, ~MCR_RTS);
				return OK;
			case TIOCMGET:		/* get all modem bits */
				disable(ps);
				uart_msr = __inb(uartp + U_MCR);
				uart_mcr = __inb(uartp + U_MSR);
				restore(ps);
				*((int *)arg) = 0;
				/* translate the bits from the UART to the ioctl */
				for (mapper = mcr_map; mapper->ioctl_bit != -1; mapper++) {
					if (uart_mcr & mapper->uart_bit) {
						*((int *)arg) |= mapper->ioctl_bit;
					}
				}
				for (mapper = msr_map; mapper->ioctl_bit != -1; mapper++) {
					if (uart_msr & mapper->uart_bit) {
						*((int *)arg) |= mapper->ioctl_bit;
					}
				}
				return OK;
			case TIOCMBIS:		/* turn on (bit-set) specific modem bits */
				disable(ps);
				__port_orb(uartp + U_MCR, ctibh_ioctl2mcr(*(int *)arg));
				__port_orb(uartp + U_MSR, ctibh_ioctl2msr(*(int *)arg));
				restore(ps);
				return OK;
			case TIOCMBIC:		/* turn off (bit-clear) specific modem bits */
				/* This request is called from an interrupt via tmgr_rx() */
				__port_andb(uartp + U_MCR, ~ctibh_ioctl2mcr(*(int *)arg));
				__port_andb(uartp + U_MSR, ~ctibh_ioctl2msr(*(int *)arg));
				restore(ps);
				return OK;
			case TIOCMSET:		/* set all modem bits */
				disable(ps);
				__outb(uartp + U_MCR, ctibh_ioctl2mcr(*(int *)arg));
				__outb(uartp + U_MSR, ctibh_ioctl2msr(*(int *)arg));
				restore(ps);
				return OK;
			case TIOCSETP:
			case OTIOCSETP:
				ctibh_xmit_quies(uartp);
				/* fall through */
			case TIOCSETN:
			case OTIOCSETN:
				ctibh_seth(s, (struct sgttyb *)arg, p, 0, tflags);
				break;
			case TCSETAW:
			case TCSETAF:
				ctibh_xmit_quies(uartp);
				/* fall through */
			case TCSETA:
				ctibh_seth_pv(s, (struct termios *)arg, p, 1);
				break;
			case TCSETAWP:	/* posix tcsetattr() */
			case TCSETAFP:	/* posix tcsetattr() */
				ctibh_xmit_quies(uartp);
				/* fall through */
			case TCSETAP:	/* posix tcsetattr() */
				ctibh_seth_pv(s, (struct termios *)arg, p, 0);
				break;
			case TIOCDRAIN:
				ctibh_xmit_quies(uartp);
				break;
			case TIOCSBRK:
				disable(ps);
				__port_orb(uartp + U_LCR, LCR_BRK);
				restore(ps);
				break;
			case TIOCCBRK:
				disable(ps);
				__port_andb(uartp + U_LCR, ~LCR_BRK);
				restore(ps);
				break;
			default:
				break;
		}
	} else if (out == SYSERR && pgeterr() == EINVAL
			&& (com == TIOKKPRINTF || com == TIOSKDB)) {
		seterr(0);
		switch (com) {
		case TIOKKPRINTF:
			kkprintf_hook(f->dev, uartp, ctibh_kputc);
			break;
		case TIOSKDB:
			if (!skdb_entry_from_irq) {
				seterr(EAGAIN);
				debug(("%s: leave\n", __FUNCTION__));
				return SYSERR;
			}
			ctibh_skdb_hook(f->dev, uartp);
			if (arg && *(int *)arg)
				skdb_key_serial = *(int *)arg;
			skdb_port_serial = uartp;
			break;
		}
		out = OK;
	}

	debug(("%s: leave\n", __FUNCTION__));
	return out;
}

int
ctibhselect(
	struct ctibh_static*  p,
	struct file*          f,
	int                   com,
	struct sel*           se
	)
{
	int  minor_no;
	int  ret;

	debug(("%s: enter (0x%X,0x%X,0x%X,0x%X)\n", __FUNCTION__, p, f, com, se));

	minor_no = minor(f->dev);
	if((minor_no < 0) || (minor_no >= p->num_ports)) {
		debug(("%s: minor number out of range (%d)\n", __FUNCTION__, minor_no));
		return(SYSERR);
	}

	ret = tmgr_select(&(p->ports[minor_no].channel), f, com, se);
	debug(("%s: leave\n", __FUNCTION__));
	return ret;
}

/*
	Busy polling putc() for SKDB and kkprintf
	(This function is actually kputc() in the kernel)
	static void ctibh_kputc(int c, int uport)
*/

/*
	Busy polling getc() for SKDB
*/
static int
ctibh_kgetc(
	int  uport
	)
{
	int  data;

#if defined(__CTIBH_SKDB_BUFF__)
	if (skdb_tty_head == skdb_tty_tail) {
#endif
		while (!(__inb(uport + U_LSR) & LSR_DR)) { ; }
		data = __inb(uport + U_RBR);
#if defined(__CTIBH_SKDB_BUFF__)
	}
	else {
		data = skdb_tty_buff[skdb_tty_tail++];
		skdb_tty_tail %= 80;
	}
#endif
	return data;
}

static int
ctibh_readhotkey(
	__UNUSED_ARG(int  uport)
	)
{
	return skdb_key_serial;
}

static void
ctibh_sethotkey(
	int  key,
	__UNUSED_ARG(int  uport)
	)
{
	skdb_key_serial = key;
}

static char *
ctibh_keyname(
	__UNUSED_ARG(int  key),
	__UNUSED_ARG(int  port)
	)
{
	return NULL;	/* SKDB will interpret as default ASCII key */
}


/*
	SKDB function vector
*/
static struct skdb_vector ctibh_skfvec = {
	ctibh_kputc,
	ctibh_kgetc,
	ctibh_kgetc,
	ctibh_readhotkey,
	ctibh_sethotkey,
	ctibh_keyname,
	NULL
};

/*
	SKDB entry from tty
*/
void
ctibh_skdb_hook(
	int  dev,
	int  port
	)
{
	skdb_hook(dev, port, &ctibh_skfvec);
}


static void
ctibh_skdb(
	struct ctibh_pstatic*  p
	)
{
	ctibh_skdb_hook(0, p->uartp);
	skdb_entry_from_irq(currtptr->iecp, -1, currtptr->islevel);
}


EP_STATIC struct dldd entry_points = {
	ctibhopen,
	ctibhclose,
	ctibhread,
	ctibhwrite,
	ctibhselect,
	ctibhioctl,
	ctibhinstall,
	ctibhuninstall,
	0
};

